from . import (
    alexnet,
    convnext,
    densenet,
    efficientnet,
    googlenet,
    mobilenetv2,
    mobilenetv3,
    regnet,
    resnet,
    shufflenetv2,
    squeezenet,
    swin,
    vgg,
    vit,
)
